/*-------------------------------------------------------------
 *  Copyright (c) 2011-2013, MSS ComBlock. All rights reserved.
 *
 * 	Filename: RF.c
 *  Last modified: 3/9/13
 *  Authors:
 *  	Alain Zarembowitch / MSS ComBlock
 *  Version: 0
 *  Inheritance: CB3011
 *  Description: RF components driver
 * -------------------------------------------------------------*/
#include "main.h"
#include "rf.h"
#include "fpga.h"


void DAC_init(void) {
	// initialize internal 10-bit DAC
	long a;

	// 1. to enable the DAC, output must be selected to appear on the related pin p0.26
	PINSEL1 = (PINSEL1 & 0xFFCFFFFF) | 0x00200000;

	// 2. peripheral clock selection for DAC.
	// select the processor clock CCLK/4 (default)
	PCLKSEL1 = PCLKSEL1 & 0xFF3FFFFF;

	// 3. enable double buffering, no interrupts, 1us refresh
	DACCTRL = 0x00000006;
	DACCNTVAL = 30; // 1us @120MHz/4
}

void DAC(long value) {
	// output is value * Vrefp / 1024
	// 1us settling time -> 1 MHz maximum update rate
	DACR = (value << 6) & 0x0000FFC0;
}

void PWM_init(void) {
	// initialized P1[18] = PWM 1.1 as a fake 10-bit DAC (externally filtered)
	// initialized P1[20] = PWM 1.2 as a fake 10-bit DAC (externally filtered)
	// initialized P1[23] = PWM 1.4 as a fake 10-bit DAC (externally filtered)
	// PWM1IR interrupt register ignored (no interrupts)
	PCONP |= (1 << 6); // PWM1 power ON
	// PCLKSEL0: by default peripheral clock is CCLK/4 = 30 MHz
	PWM1TCR = 0x00000002; // Reset counter
	PWM1PR = 0;
	PWM1PC = 0;
	PINSEL3 = (PINSEL3 & 0xFFFF3CCF) | 0x00008220;
	// p1[18,20,23] = PWM
	PINMODE3 = (PINMODE3 & 0xFFFF3CCF) | 0x00008220;
	// P1[18,20,23] PWM has no PU/PD
	PWM1MR0 = 1024; // controls the PWM cycle rate. Here 1024 peripheral clocks
	PWM1MR1 = 0; // initial PWM value
	PWM1MR2 = 0; // initial PWM value
	PWM1MR4 = 0; // initial PWM value
	// PWM Match Control Register
	// Match 0 is configured to reset the timer/counter when a match event occurs.
	PWM1MCR = 0x00000002; // reset counter when match MR0
	// PWM Control Register
	PWM1PCR = 0x00001600; // all single edge, PWM1,2,4 outputs enabled
	// PWM Timer Control Register
	PWM1TCR = 0x00000009; // clear counter,  enable PWM mode, enable counting
	// output is value * 3.3V / 1024 after filtering
	PWM1MR0 = 1024; // controls the PWM cycle rate. Here 1024 peripheral clocks
	// PWM Latch Enable Register
	PWM1LER = 0x00000017; // Enable PWM Match 0,1,2,4 Latch
}

void PWM11(long value) {
	// output is value * 3.3V / 1024 after filtering
	PWM1MR1 = value & 0x000003FF; // controls the PWM edge position
	// PWM Latch Enable Register
	PWM1LER = 0x00000017; // Enable PWM Match 0,1,2,4 Latch
}

void PWM12(long value) {
	// output is value * 3.3V / 1024 after filtering
	PWM1MR2 = value & 0x000003FF; // controls the PWM edge position
	// PWM Latch Enable Register
	PWM1LER = 0x00000017; // Enable PWM Match 0,1,2,4 Latch
}

void PWM14(long value) {
	// output is value * 3.3V / 1024 after filtering
	PWM1MR4 = value & 0x000003FF; // controls the PWM edge position
	// PWM Latch Enable Register
	PWM1LER = 0x00000017; // Enable PWM Match 0,1,2,4 Latch
}

void rx_AGC_poll(void){
	// read FPGA AGC_DAC(11:0) in status registers REG10/9 (max value for min gain).
	// then apply gain to internal IF and external RF variable gain amplifiers if enabled.
	unsigned int c;

	// dummy read at status register 8 to latch other multi-byte fields
	c = fpga_getByte(8);

	if(regv[0] & 0x04){
		// IF1 AGC enabled
		// close the AGC loop from status (SREG10/9) to PWM11
		// change definition from [0-4095] (max/min gain) to [1023-0](max/min gain) at PIN diode attenuator
		c = fpga_getByte(10);
		c = (c << 8) & 0x00000F00;
		c = c | (fpga_getByte(9) & 0x00FF);
		c >>= 2;
		c = 1023 - c;
		PWM11(c);
	}
	if(regv[0] & 0x01){
		// external RF AGC enabled
		// close the AGC loop from status (SREG10/9) to control(REG4/3)
		// change definition from [0-4095] (max/min gain) to [0-1023](max/min gain)
		c = fpga_getByte(10);
		c = (c << 8) & 0x00000F00;
		c = c | (fpga_getByte(9) & 0x00FF);
		c >>= 2;
		PWM12(c);
	}
}
/*
void RF_AGC_poll(void) {
	// 3 AGC modes as set in REG69(7:6)
	unsigned char agc_loop, a;
	agc_loop = (regv[69] & 0xC0) >> 6;

	// check level/power sensors for reporting status (even when AGC loop is open) 3/9/13
	ADC_RF_POWER_DET();
	if (agc_loop == 0) {
		// open loop. Gain is fixed.
		return;
		// RF AGC loop algorithm relying on RF power detection (prevent saturation at
		// RF mixer) and peak level after the A/D converter (prevent saturation at the A/D converter)
	} else if ((agc_loop == 1) || (agc_loop == 2)) {
		// RF AGC loop under this processor control
		// read peak detect status from FPGA (SREG1)
		// The peak amplitude is detected over a 256-sample window at the ADC output = 64 signal periods.
		a = fpga_getByte(1);
		// read RF power detect rf_power_det (done periodically in main loop)
		// new: threshold reduced to prevent saturation (wideband noise + signal)
		if ((rf_power_det > 0x0B40) || (a > 0x00C0)) {
			// getting close to saturation, either at the RF mixer or at the A/D converter. Decrease gain
			RF_gain_too_large_duration++;
			RF_gain_too_small_duration = 0; // reset the time elapsed since last peak detect > 0x80
			if (RF_gain_too_large_duration > 16) { // agc reaction time implied here
				RF_gain_too_large_duration = 0; // reset counter
				RF_gain_decrement();
			}
		} else {
			// slow pressure to increase gain.
			RF_gain_too_small_duration++; // increment the time elapsed since last saturation detect
			if (RF_gain_too_small_duration > 16) { // agc reaction time implied here
				// gain is consistently too small.
				RF_gain_increment();
				RF_gain_too_large_duration = 0; // reset counter
				RF_gain_too_small_duration = 0; // reset the time elapsed since last saturation detect
			}
		}
	}
}

void BB_AGC_poll(void) {
	// 3 AGC modes as set in REG69(7:6)
	unsigned char agc_loop, ai, aq;
	agc_loop = (regv[69] & 0xC0) >> 6;

	if (agc_loop == 0) {
		// open loop. Gain is fixed.
		return;
		// Baseband AGC loop algorithm relies on detecting saturation or near saturation at the
		// output samples.
	} else if (agc_loop == 1) {
		// Baseband AGC loop under this processor control
		// read output samples MSBs from FPGA (SREG2/SREG3)
		ai = fpga_getByte(2);
		aq = fpga_getByte(3);
		if((ai > 0x00F8) || (ai < 0x0008) || (aq > 0x00F8) || (aq < 0x0008)){
			// Near saturation detected. Reduce gain
			if(BB_gain > 16)
				BB_gain -= 16;
		} else {
			// slow pressure to increase the gain
			if(BB_gain < 65535)
				BB_gain++;
		}
		// set digital gain
		setBBGain();

	} else if (agc_loop == 2) {
		// external AGC loop
		// read ADC (converting the AGC_IN signal from the 40-pin connector to a 12-bit value)
		ADC_AGC_IN();
		// flip gain control value (ComBlocks AGC rule: 0 is for maximum gain)
		BB_gain = 65535 - (agc_in << 4);	// agc_in range 0 - 4095
		// set digital gain
		setBBGain();
	}
}


//long pinDiodeAttenuatorCompensation(long gain){
//	// attempt at avoiding instabilities
//	// function of center frequency fc (global)
//	// input range 0 - 1023
//
//	long revised_gain;
//	revised_gain = gain;
//
//	if(fc < 200000000){
//		// below 200 MHz, avoid very large attenuations (Vc < 0.55V) which are destabilizing
//		if(gain < 129)
//			revised_gain = 129;
//	} else if (fc < 500000000){
//		// below 500 MHz, avoid very large attenuations (Vc < 0.5V) which are destabilizing
//		if(gain < 108)
//			revised_gain = 108;
//	} else if (fc > 2000000000){
//		// above 2 GHz, avoid too much gain (> 2.5V) because it can create instabilities
//		if(gain > 538)
//			revised_gain = 538;
//	}
//	return revised_gain;
//}

void ADC_init(void) {
	// ADC initialization
	PCONP |= 0x00001000; // power ADC subsystem
	// PCLKSEL0: by default peripheral clock is CCLK/4 = 30 MHz
	// A/D Control Register: inputs are AD0[2] and [5], ADC clock is 10 MHz, burst=1, operational,
	AD0CR = 0x00210224;
}

void ADC_AGC_IN(void) {
	// returns the last value read on AD0[5] = AGC_IN from 40-pin connector, low-pass filtered
	unsigned long a;
	a = AD0DR5;
	if (a & 0x80000000) {
		// done bit high. update agc_in
		agc_in = (a >> 4) & 0x00000FFF; // 12-bit ADC measurement
	}
}

void ADC_RF_POWER_DET(void) {
	// returns the last value read on AD0[2] = RF power detected from the RF mixer
	unsigned long a;
	a = AD0DR2;
	if (a & 0x80000000) {
		// done bit high. update rf_power_det
		rf_power_det = (a >> 4) & 0x00000FFF; // 12-bit ADC measurement
	}
}
*/
