/*-------------------------------------------------------------
 *  Copyright (c) 2010-2012, MSS ComBlock. All rights reserved.
 *
 * 	Filename: uart.c
 *  Last modified: 8/19/12
 *  Authors:
 *  	Alain Zarembowitch / MSS ComBlock
 *  Version: 4
 *  Inheritance: N/A
 *  Description: UART0 and UART3 configurations.
 *
 *  Dependencies/Pre-requisite:
 *  1. power up the UART0 & UART3 peripherals using PCONP (initConfig.c)
 *  2. configure the clocks in clock_power_profile(). Key assumption is that the CPU clock is CCLK = 120 MHz
 *
 *  Rev 1 8/22/10
 *  Rev 2 9/30/10 AZ
 *  Changed UART from one to two stop bits, otherwise will not work with older (COM-2001, etc) ComBlocks.
 *  Rev  3 4/17/11 AZ
 *  Optimize USB_ARM_mcrx for speed. Copy USB rx data directly to rxLine.buf instead of doing two successive copies.
 *  Solved flash programming reliability bug (it took too long to copy data before the next 64B bulk OUT packet)
 *
 *  Rev4 8/19/12 AZ
 *  Bug correction: in USB_FPGA_rxpoll()+uart_rxLineAppend()
 *  Changed ifs order to avoid M&C deadlock at power up
 -------------------------------------------------------------*/
#include "main.h"
#include "uart.h"
#include "vectors.h"
#include "fpga.h"
#include "lpc_types.h"
#include "usb.h"
#include "usbhw.h"

long USB_ARM_EP82bufempty = 1;

void uart_init() {
	// configure UART0 and UART3 for 115.2 Kbaud, 8-bit, no parity, one stop bit
	// Auto-baud feature is not used.
	// Key assumption: PCLK = CCLK/4 = 30 MHz.

	// copied from com-1800???
//	PCONP |= 0x00000008;	// enable UART0
//	PCONP |= 0x02000000;	// enable UART3

	// UARTn Line Control Register (UnLCR)
	// 8-bit, no parity, two stop bits. enable DLAB1 before changing divisor latches
	U0LCR = 0x00000087;
	U3LCR = 0x00000087;

	// Baud rate configuration 	115.2Kbaud => DL = 10, Div/Mul = 5/8
	// UARTn Divisor Latch registers (UnDLL, UnDLM)
	U0DLL = 10; // LSB
	U0DLM = 0; // MSB
	U3DLL = 10;
	U3DLM = 0;
	// UARTn Fractional Divider Register (UnFDR)
	U0FDR = 0x00000085;
	U3FDR = 0x00000085;
	// enable the tx/rx FIFOs
	U0FCR |= 0x00000001;
	U3FCR |= 0x00000001;
	// disable access to divisor latches. Enable access to U0/3IER
	U0LCR &= ~0x00000080;
	U3LCR &= ~0x00000080;
	// disable UART interrupts
	// UARTn Interrupt Enable Register (UnIER)
	U0IER = 0x00000000;
	U3IER = 0x00000000;

	return;
}

/*
 WEAK void UART0_IRQHandler(void){
 // UART0 interrupt handler
 int c;

 // find the interrupt source
 c = U0IIR & 0x0000000E;
 if((c == 0x00000004) || (c == 0x0000000C)){
 // 2a - Receive Data Available (RDA) or
 // 2b - Character Time-out Indicator (CTI).
 // transfer data from the rx FIFO to the command line handler
 // stop when rx FIFO is empty or when a full command line is received or when line is full, whichever comes first.
 uart_rxLineAppend(0);
 } else if(c == 0x00000002){
 // 3 - THRE Interrupt
 uart_txLineSend();
 } else if(c == 0x00000006){
 // 1 - RLS interrupt
 c = U0LSR;	// clear the interrupt
 return;
 }
 // clear all interrupt for UART0
 //ICPR0 = 0x00000020;
 return;
 }

 WEAK void UART3_IRQHandler(void){
 // UART3 interrupt handler
 int c;

 // find the interrupt source
 c = U3IIR & 0x0000000E;

 // find the interrupt source
 if((c == 0x00000004) || (c == 0x0000000C)){
 // 2a - Receive Data Available (RDA) or
 // 2b - Character Time-out Indicator (CTI).
 // transfer data from the rx FIFO to the command line handler
 // stop when rx FIFO is empty or when a full command line is received or when line is full, whichever comes first.
 uart_rxLineAppend(3);
 } else if(c == 0x00000002){
 // 3 - THRE Interrupt
 uart_txLineSend();
 } else if(c == 0x00000006){
 // 1 - RLS interrupt
 c = U3LSR;	// clear the interrupt
 return;
 }
 // clear all interrupt for UART3
 //ICPR0 = 0x00000100;
 return;
 }
 */

long uart_rxpoll(void) {
	// polling UART0/3 for rx data.
	long retval;

	retval = 0;
	if (U0LSR & 0x00000001) {
		//if (U0FIFOLVL & 0x0000000F) {
		// data in UART0 rx FIFO. transfer data.
		retval = uart_rxLineAppend(0);
	} else if (U3LSR & 0x00000001) {
		//	} else if (U3FIFOLVL & 0x0000000F) {
		// data in UART3 rx FIFO. transfer data.
		retval = uart_rxLineAppend(3);
	}

	return retval;
}

// Note: a similar function exists to transfer data from a USB source to rxLine.
// Usage: process the received message when the msg_complete flag is true.
// When done processing the received message, move the rptr = wptr.
long uart_rxLineAppend(long uart_source) {
	// transfer data from various UART rx FIFOs to a global buffer: rxLine.buf[]
	// stop when rx FIFO is empty or when a full command line is received or when line is full, whichever comes first.
	// Add a \0 at the end of the rx message so as to allow the use of strlen()
	char c;

	if (uart_source == 0) {
		// read from UART0
		if ((rxLine.msg_complete) && (rxLine.rptr == 0)) {
			// We have not yet processed the previous message. ignore (dump) new incoming characters
			while (U0LSR & 0x00000001) {
				// The UARTn receiver FIFO is not empty.
				c = (char) (U0RBR);
			}
			return -UART_ERXTOOFAST;
		}

		while (U0LSR & 0x00000001) {
			// The UARTn receiver FIFO is not empty.
			c = (char) (U0RBR);
			// changed order 8/19/12 az
			if (c == '@') {
				// new start of line. rewind read/write pointers
				rxLine.msg_complete = 0;
				rxLine.wptr = 0;
				rxLine.rptr = 0;
				rxLine.portNo = 4; // attribute message source to left connector
			}
			if (rxLine.wptr >= sizeof(rxLine.buf)) {
				// rxLine is full. can't write another character (not even a 0 for end of string)
				rxLine.msg_complete = 1;
				return -UART_ERXFULL;
			}
			rxLine.buf[rxLine.wptr++] = c; // transfer character
			if (c == '\n') {
				// end of line
				if (rxLine.wptr < sizeof(rxLine.buf)) {
					// there is room for an end of string character 0
					rxLine.buf[rxLine.wptr] = 0;
				}
				rxLine.msg_complete = 1;
				break;
			}
		}
		return UART_OK;
	} else if (uart_source == 3) {
		// read from UART3
		if ((rxLine.msg_complete) && (rxLine.rptr == 0)) {
			// We have not yet processed the previous message. ignore (dump) new incoming characters
			while (U3LSR & 0x00000001) {
				// The UARTn receiver FIFO is not empty.
				c = (char) (U3RBR);
			}
			return -UART_ERXTOOFAST;
		}

		while (U3LSR & 0x00000001) {
			// The UARTn receiver FIFO is not empty.
			c = (char) (U3RBR);
			// changed order 8/19/12 az
			if (c == '@') {
				// new start of line. rewind read/write pointers
				rxLine.msg_complete = 0;
				rxLine.wptr = 0;
				rxLine.rptr = 0;
				rxLine.portNo = 2; // attribute message source to right connector
			}
			if (rxLine.wptr >= sizeof(rxLine.buf)) {
				// rxLine is full. can't write another character (not even a 0 for end of string)
				rxLine.buf[sizeof(rxLine.buf) - 1] = 0; // force end message with end of string character 0
				rxLine.msg_complete = 1;
				return -UART_ERXFULL;
			}
			rxLine.buf[rxLine.wptr++] = c; // transfer character
			if (c == '\n') {
				// end of line
				if (rxLine.wptr < sizeof(rxLine.buf)) {
					// there is room for an end of string character 0
					rxLine.buf[rxLine.wptr] = 0;
				}
				rxLine.msg_complete = 1;
				break;
			}
		}
		return UART_OK;
	}
	return -UART_EDATAIN;
}

// usage:
// 1. write message in txLine.buf, update write pointer wptr, specify the destination port portNo.
// 2. when the message assembly is complete, reset the read pointer to 0 and and clear the complete flag
// 3. call uart_txLineSend once.
// 4. each polling call automatically transfers data from txLine.buf to the output UART, 16 bytes at a time.
// 5. when the message is completely transmitted, the complete flag is set in txLine.complete
long uart_txpoll(void) {
	// push tx line to the designated tx UART.

	if ((txLine.wptr <= txLine.rptr) || (txLine.msg_complete == 0)) {
		// nothing to send.
		return UART_OK;
	} else if (txLine.portNo == 4) {
		// compute timeout for message transmission time
		timer[3] = 4; // 30-40ms timeout (longest message is  23.8ms
		while (1) {
			if (timer[3] == 0) {
				// timeout
				txLine.msg_complete = 0; // ready for next tx message
				txLine.rptr = 0;
				return -UART_TXTIMEOUT;
			} else if (txLine.rptr == txLine.wptr) {
				// transfer is complete.
				txLine.msg_complete = 0; // ready for next tx message
				txLine.rptr = 0;
				return UART_OK;
			} else if ((U0FIFOLVL & 0x00000F00) != 0x00000F00) {
				// UART0 tx FIFO is not full and there is still data to transmit
				U0THR = txLine.buf[txLine.rptr++];
			}
		}
	} else if (txLine.portNo == 2) {
		timer[3] = 4; // 30-40ms timeout (longest message is  23.8ms
		while (1) {
			if (timer[3] == 0) {
				// timeout
				txLine.msg_complete = 0; // ready for next tx message
				txLine.rptr = 0;
				return -UART_TXTIMEOUT;
			} else if (txLine.rptr == txLine.wptr) {
				// transfer is complete.
				txLine.msg_complete = 0; // ready for next tx message
				txLine.rptr = 0;
				return UART_OK;
			} else if ((U3FIFOLVL & 0x00000F00) != 0x00000F00) {
				// UART3 tx FIFO is not full and there is still data to transmit
				U3THR = txLine.buf[txLine.rptr++];
			}
		}
	}
	return UART_OK;
}

// not quite UART but close.
// serial communication (USB or LAN) through FPGA
long USB_FPGA_rxpoll(void) {
	// non-blocking attempt to read incoming messages from the USB via the FPGA.
	char c;

	// read from USB2.0 via FPGA
	if ((rxLine.msg_complete) && (rxLine.rptr == 0)) {
		// We have not yet processed the previous message. ignore (dump) new incoming characters
		while (!(fpga_getByte(255) & 0x08)) {
			// RX_USB_BUFFER is not empty
			c = fpga_getByte(254);
		}
		return -UART_ERXTOOFAST;
	}

	while (!(fpga_getByte(255) & 0x08)) {
		// RX_USB_BUFFER is not empty
		c = fpga_getByte(254);
		// changed order 8/19/12 az
		if (c == '@') {
			// new start of line. rewind read/write pointers
			rxLine.msg_complete = 0;
			rxLine.wptr = 0;
			rxLine.rptr = 0;
			rxLine.portNo = 3; // attribute message source to FPGA USB2.0 interface

		}
		if (rxLine.wptr >= sizeof(rxLine.buf)) {
			// rxLine is full. can't write another character (not even a 0 for end of string)
			// Note 12/20/15: important, keep. Maybe needed in com-1800?
			rxLine.msg_complete = 1;
			return -UART_ERXFULL;
		}
		rxLine.buf[rxLine.wptr++] = c; // transfer character
		if (c == '\n') {
			// end of line
			if (rxLine.wptr < sizeof(rxLine.buf)) {
				// there is room for an end of string character 0
				rxLine.buf[rxLine.wptr] = 0;
			}
			rxLine.msg_complete = 1;
			break;
		}
	}
	return UART_OK;
}

long USB_FPGA_txpoll(void) {
	//  send tx message from txLine to USB via FPGA (portNo 3)

	if ((txLine.wptr <= txLine.rptr) || (txLine.msg_complete == 0)) {
		// nothing to send.
		return UART_OK;
	}
	// compute timeout for message transmission time
	timer[3] = 2; // 10-20ms timeout
	while (1) {
		if (timer[3] == 0) {
			// timeout
			txLine.msg_complete = 0; // ready for next tx message
			txLine.rptr = 0;
			return -UART_TXTIMEOUT;
		} else if (txLine.rptr == txLine.wptr) {
			// transfer is complete.
			txLine.msg_complete = 0; // ready for next tx message
			txLine.rptr = 0;
			return UART_OK;
		} else {
			// There is still data to transmit
			// No tx flow control here. USB is always faster and has more elastic buffer than we need.
			fpga_setByte(255, txLine.buf[txLine.rptr++]);
		}
	}
	return UART_OK;
}

// not quite UART but close.
// USB serial communication through LPC17xx ARM
// Optimize for speed.
long USB_ARM_mcrx(void) {
	// callback upon receiving a BulkOut event (64 Bytes max) from the USB DEVelopment port via the LPC17xx ARM
	char c;
	unsigned long len; /* Bulk In/Out Length */
	unsigned long i, imin;
	unsigned long maxLen;

//	FIO1SET = 0x00000002;	// test
	// read from USB2.0 via ARM
	if ((rxLine.msg_complete) && (rxLine.rptr == 0)) {
		// We have not yet processed the previous message. discard new incoming characters
		// clear USB rx buffer
		USB_ClearRxEP(0x02); // discard rx data
		return -UART_ERXTOOFAST;
	} else if (rxLine.rptr == rxLine.wptr) {
		// we are about to receive a new message
		// reset wptr
		rxLine.wptr = 0;
		rxLine.msg_complete = 0;
	}

	maxLen = sizeof(rxLine.buf) - rxLine.wptr; // make sure we don't copy USB rx data past the end of buffer.
	// copy USB data directly to rxLine.buf, to save time.
	len = USB_ReadEP2(0x02, &(rxLine.buf[rxLine.wptr]), maxLen); // read  endpoint 0x02 (Out direction)

	// scan new data for end of line \n
	imin = rxLine.wptr;
	rxLine.wptr += len;
	for (i = imin; i < rxLine.wptr; i++) {
		// read buffer one character at a time
		c = rxLine.buf[i];
		if (i == sizeof(rxLine.buf) - 1) {
			// buffer is full => message is complete
			rxLine.buf[i] = 0; // force an end of string
			rxLine.rptr = 0;
			rxLine.portNo = 1; // attribute message source to ARM USB2.0 interface
			rxLine.msg_complete = 1;
		} else if (c == '\n') {
			// end of line or buffer is full => message is complete
			rxLine.buf[i+1] = 0; // force an end of string
			rxLine.rptr = 0;
			rxLine.portNo = 1; // attribute message source to ARM USB2.0 interface
			rxLine.msg_complete = 1;
		}
	}
	// All control messages should start with @. It does not hurt to verify though.
	if ((rxLine.wptr != 0) && (rxLine.buf[0] != '@')) {
		// abnormal condition. message should start with @.
		// find first @
		for (i = 0; i < rxLine.wptr; i++) {
			c = rxLine.buf[i];
			if (c == '@') {
				// found the start. Shift by i positions to the left
				rxLine.wptr -= i;
				memcpy(&(rxLine.buf[0]), &(rxLine.buf[i]), rxLine.wptr);
				break;
			}
		}
	}
//	FIO1CLR = 0x00000002;	// test
	return UART_OK;
}

long USB_ARM_mctx(void) {
	//  send tx message from txLine to USB via ARM (portNo 1)  64 bytes at a time.
	long l;

	if ((txLine.wptr <= txLine.rptr) || (txLine.msg_complete == 0)) {
		// nothing to send.
		return UART_OK;
	}
	// compute timeout for message transmission time
	timer[3] = 2; // 10-20ms timeout
	while (1) {
		if (timer[3] == 0) {
			// timeout
			USB_ARM_EP82bufempty = 1; // not sure why we timed out. Could be the host not reading. In this case, discard message to recover.
			txLine.msg_complete = 0; // ready for next tx message
			txLine.rptr = 0;
			return -UART_TXTIMEOUT;
		} else if (txLine.rptr == txLine.wptr) {
			// transfer is complete.
			txLine.msg_complete = 0; // ready for next tx message
			txLine.rptr = 0;
			return UART_OK;
		} else if (USB_ARM_EP82bufempty) {
			// There is still data to transmit and endpoint buffer 0x82 is empty
			l = txLine.wptr - txLine.rptr;
			if (l > 64)
				l = 64; // maximum packet size Bulk IN
			// write data to the endpoint buffer
			USB_ARM_EP82bufempty = 0;
			USB_WriteEP(0x82, (unsigned char*) (&txLine.buf[txLine.rptr]), l);
			txLine.rptr += l;
		}
	}
	return UART_OK;
}

void USB_ARM_mctxint(void) {
	// callback routine when Bulk IN packet was transmitted
	// Generally indicates that it is OK to send the next packet whenever
	USB_ARM_EP82bufempty = 1;
	return;
}
